package sk.fourq.otaupdate;

import android.Manifest;
import android.annotation.TargetApi;
import android.app.DownloadManager;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageManager;
import android.database.Cursor;
import android.net.Uri;
import android.os.Build;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.os.Message;
import android.util.Log;

import org.json.JSONObject;

import io.flutter.plugin.common.EventChannel;
import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.PluginRegistry;
import io.flutter.plugin.common.PluginRegistry.Registrar;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.ref.WeakReference;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.util.Arrays;

import androidx.core.content.FileProvider;
import androidx.core.content.ContextCompat;
import androidx.core.app.ActivityCompat;

/**
 * OtaUpdatePlugin
 */
@TargetApi(Build.VERSION_CODES.M)
public class OtaUpdatePlugin implements EventChannel.StreamHandler, PluginRegistry.RequestPermissionsResultListener, MethodChannel.MethodCallHandler {
    private UIhandler ihandler;
    @Override
    public void onMethodCall(MethodCall call, MethodChannel.Result result) {
        Class<?> clazz = this.getClass();
        try {
            if (call.arguments != null) {
                Method method = clazz.getDeclaredMethod(call.method, new Class<?>[] { HashMap.class,MethodChannel.Result.class });
                method.setAccessible(true);
                Object r = method.invoke(this, call.arguments, result);
            } else {
                Method method = clazz.getDeclaredMethod(call.method, MethodChannel.Result.class );
                method.setAccessible(true);
                Object r = method.invoke(this,result);
            }
        } catch (Exception e) {
            e.printStackTrace();
            result.notImplemented();
        }
    }

    enum OtaStatus {
        DOWNLOADING, INSTALLING, ALREADY_RUNNING_ERROR, PERMISSION_NOT_GRANTED_ERROR, INTERNAL_ERROR
    }

    private final Registrar registrar;
    private EventChannel.EventSink progressSink;
    private String downloadUrl;
    private String downloadName;
    private MethodChannel.Result _result;
    private String destinationRoot = Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_DOWNLOADS) + "";

    private static final String TAG = "FLUTTER OTA";

    private OtaUpdatePlugin(Registrar registrar) {
        this.registrar = registrar;
        ihandler = new UIhandler(this);
    }

    /**
     * Plugin registration.
     */
    public static void registerWith(Registrar registrar) {
        OtaUpdatePlugin plugin = new OtaUpdatePlugin(registrar);
        final EventChannel progressChannel = new EventChannel(registrar.messenger(), "sk.fourq.ota_update");
        progressChannel.setStreamHandler(plugin);
        registrar.addRequestPermissionsResultListener(plugin);

        final MethodChannel channel = new MethodChannel(registrar.messenger(), "ota_update");
        channel.setMethodCallHandler(plugin);
    }

    // 安装apk
    private void installApk(final HashMap jsonData, final MethodChannel.Result result){
        final Context context = (registrar.activity() != null) ? registrar.activity() : registrar.context();
        Intent intent;
        downloadName = jsonData.get("name").toString();
        if(downloadName == "" || downloadName == null){
            downloadName = "ordo";
        }
        // 实际路径
        String scrPath = destinationRoot + "/" + downloadName + ".apk";
        final Uri fileUri = Uri.parse("file://" + scrPath);
        File file = new File(scrPath);
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N) {
            //AUTHORITY NEEDS TO BE THE SAME ALSO IN MANIFEST
            String packageName = context.getApplicationContext().getPackageName();
            String authority =  new StringBuilder(packageName).append(".provider").toString();
            Uri apkUri = FileProvider.getUriForFile(context, authority, file);
            intent = new Intent(Intent.ACTION_INSTALL_PACKAGE);
            intent.setData(apkUri);
            intent.setFlags(Intent.FLAG_GRANT_READ_URI_PERMISSION);
        } else {
            intent = new Intent(Intent.ACTION_VIEW);
            intent.setDataAndType(fileUri, "application/vnd.android.package-archive");
            intent.setFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
        }
        context.startActivity(intent);
    }

    // 是否已存在apk
    private void exists(final HashMap jsonData, final MethodChannel.Result result){
        _result = result;
        downloadName = jsonData.get("name").toString();
        if (PackageManager.PERMISSION_GRANTED == ContextCompat.checkSelfPermission(registrar.context(), Manifest.permission.READ_EXTERNAL_STORAGE)) {
            HashMap rt=new HashMap();
            rt.put("status",true);
            rt.put("data",exist());
            result.success(rt);
        } else {
            String[] permissions = {
                    Manifest.permission.READ_EXTERNAL_STORAGE
            };
            ActivityCompat.requestPermissions(registrar.activity(), permissions, 1);
        }
    }
    private boolean exist(){
        if(downloadName == "" || downloadName == null){
            downloadName = "ordo";
        }
        // 实际路径
        String scrPath = destinationRoot + "/" + downloadName + ".apk";
        File file = new File(scrPath);
        return file.exists();
    }

    @Override
    public void onListen(Object arguments, EventChannel.EventSink events) {
        if (progressSink != null) {
            progressSink.error("" + OtaStatus.ALREADY_RUNNING_ERROR.ordinal(), "Method call was cancelled. One method call is already running", null);
            return;
        }
        progressSink = events;
        downloadUrl = ((Map)arguments).get("url").toString();
        if(((Map) arguments).containsKey("name")){
            downloadName = ((Map)arguments).get("name").toString();
        }

        if (
//                PackageManager.PERMISSION_GRANTED == ContextCompat.checkSelfPermission(registrar.context(), Manifest.permission.ACCESS_WIFI_STATE) &&
//                PackageManager.PERMISSION_GRANTED == ContextCompat.checkSelfPermission(registrar.context(), Manifest.permission.ACCESS_NETWORK_STATE) &&
                PackageManager.PERMISSION_GRANTED == ContextCompat.checkSelfPermission(registrar.context(), Manifest.permission.WRITE_EXTERNAL_STORAGE)) {
            handleCall();
        } else {
            String[] permissions = {
//                    Manifest.permission.ACCESS_WIFI_STATE,
//                    Manifest.permission.ACCESS_NETWORK_STATE,
                    Manifest.permission.WRITE_EXTERNAL_STORAGE
            };
            ActivityCompat.requestPermissions(registrar.activity(), permissions, 0);
        }
    }

    @Override
    public void onCancel(Object o) {
        progressSink = null;
    }

    @Override
    public boolean onRequestPermissionsResult(int requestCode, String[] strings, int[] grantResults) {
        HashMap rt=new HashMap();
        if ((requestCode == 0 || requestCode == 1) && grantResults.length > 0) {
            for (int grantResult : grantResults) {
                if (grantResult != PackageManager.PERMISSION_GRANTED) {
                    if(requestCode == 1){
                        rt.put("status",false);
                        rt.put("msg", "无权限");
                        _result.success(rt);
                        return false;
                    }
                    progressSink.error("" + OtaStatus.PERMISSION_NOT_GRANTED_ERROR.ordinal(), null, null);
                    progressSink = null;
                    return false;
                }
            }
            if(requestCode == 1){
                rt.put("status",true);
                rt.put("data",exist());
                _result.success(rt);
                return true;
            }
            handleCall();
            return true;
        } else {
            if(requestCode == 1){
                rt.put("status",false);
                rt.put("msg", "无权限");
                _result.success(rt);
                return false;
            }
            if (progressSink != null) {
                progressSink.error("" + OtaStatus.PERMISSION_NOT_GRANTED_ERROR.ordinal(), null, null);
                progressSink = null;
            }
            return false;
        }
    }

    private void handleCall() {
        try {
            final Context context = (registrar.activity() != null) ? registrar.activity() : registrar.context();
            if(downloadName == "" || downloadName == null){
                downloadName = "ordo";
            }
            // 临时路径
            final String destination = destinationRoot + "/temp.apk";
            // 实际路径
            final String scrPath = destinationRoot + "/" + downloadName + ".apk";
            final Uri srcUri = Uri.parse("file://" + scrPath);
            final Uri fileUri = Uri.parse("file://" + destination);

            final File scrFile = new File(scrPath);
            //DELETE APK FILE IF SOME ALREADY EXISTS
            final File file = new File(destination);
            if (file.exists()) {
                if (!file.delete()) {
                    Log.e(TAG, "ERROR: unable to delete old apk file before starting OTA");
                }
            }

            //CREATE DOWNLOAD MANAGER REQUEST
            DownloadManager.Request request = new DownloadManager.Request(Uri.parse(downloadUrl));
            request.setDestinationUri(fileUri);

            //GET DOWNLOAD SERVICE AND ENQUEUE OUR DOWNLOAD REQUEST
            final DownloadManager manager = (DownloadManager) context.getSystemService(Context.DOWNLOAD_SERVICE);
            final long downloadId = manager.enqueue(request);

            //START TRACKING DOWNLOAD PROGRESS IN SEPARATE THREAD
            trackDownloadProgress(downloadId, manager);

            //REGISTER LISTENER TO KNOW WHEN DOWNLOAD IS COMPLETE
            context.registerReceiver(new BroadcastReceiver() {
                public void onReceive(Context c, Intent i) {
                    //DOWNLOAD IS COMPLETE, UNREGISTER RECEIVER AND CLOSE PROGRESS SINK
                    if (scrFile.exists()) {
                        scrFile.delete();
                    }

                    FileInputStream inputStream = null;
                    try {
                        inputStream = new FileInputStream(destination);
                        FileOutputStream outputStream = new FileOutputStream(scrPath);
                        byte[] bytes = new byte[8192];
                        int number;
                        while ((number = inputStream.read(bytes)) > 0) {
                            outputStream.write(bytes, 0, number);
                        }
                        outputStream.close();
                        inputStream.close();
                        file.delete();
                    } catch (FileNotFoundException e) {
                        e.printStackTrace();
                    } catch (IOException e) {
                        e.printStackTrace();
                    }

                    context.unregisterReceiver(this);
                    //TRIGGER APK INSTALLATION
                    Intent intent;
                    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N) {
                        //AUTHORITY NEEDS TO BE THE SAME ALSO IN MANIFEST
                        Uri apkUri;
                        String packageName = context.getApplicationContext().getPackageName();
                        String authority =  new StringBuilder(packageName).append(".provider").toString();
                        if(!scrFile.exists()){
                            apkUri = FileProvider.getUriForFile(context, authority, file);
                        }else{
                            apkUri = FileProvider.getUriForFile(context, authority, scrFile);
                        }
                        intent = new Intent(Intent.ACTION_INSTALL_PACKAGE);
                        intent.setData(apkUri);
                        intent.setFlags(Intent.FLAG_GRANT_READ_URI_PERMISSION);
                    } else {
                        intent = new Intent(Intent.ACTION_VIEW);
                        if(!scrFile.exists()){
                            intent.setDataAndType(fileUri, "application/vnd.android.package-archive");
                        }else{
                            intent.setDataAndType(srcUri, "application/vnd.android.package-archive");
                        }
                        intent.setFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
                    }
                    //SEND INSTALLING EVENT
                    progressSink.success(Arrays.asList("" + OtaStatus.INSTALLING.ordinal(), ""));
                    progressSink.endOfStream();
                    progressSink = null;
                    context.startActivity(intent);
                }
            }, new IntentFilter(DownloadManager.ACTION_DOWNLOAD_COMPLETE));
        } catch (Exception e) {
            progressSink.error("" + OtaStatus.INTERNAL_ERROR.ordinal(), e.getMessage(), null);
            progressSink = null;
            Log.e(TAG, "ERROR: " + e.getMessage(), e);
        }
    }

    private void trackDownloadProgress(final long downloadId, final DownloadManager manager) {
        new Thread(new Runnable() {
            @Override
            public void run() {
                //REPORT PROGRESS WHILE DOWNLOAD STILL RUNS
                boolean downloading = true;
                while (downloading) {
                    //QUERY CURRENT PROGRESS STATUS
                    DownloadManager.Query q = new DownloadManager.Query();
                    q.setFilterById(downloadId);
                    Cursor c = manager.query(q);
                    c.moveToFirst();
                    //PUSH THE STATUS THROUGH THE SINK
                    int bytes_downloaded = c.getInt(c.getColumnIndex(DownloadManager.COLUMN_BYTES_DOWNLOADED_SO_FAR));
                    int bytes_total = c.getInt(c.getColumnIndex(DownloadManager.COLUMN_TOTAL_SIZE_BYTES));
                    ihandler.sendEmptyMessage(1);
                    Bundle b = new Bundle();
                    b.putDouble("bytes_downloaded", bytes_downloaded);
                    b.putDouble("bytes_total", bytes_total);
                    Message msg = new Message();
                    msg.setData(b);
                    ihandler.sendMessage(msg);
//                    if (progressSink != null) {
//                        progressSink.success(Arrays.asList("" + OtaStatus.DOWNLOADING.ordinal(), "" + ((bytes_downloaded * 100) / bytes_total)));
//                    }
                    //STOP CYCLE IF DOWNLOAD IS COMPLETE
                    if (c.getInt(c.getColumnIndex(DownloadManager.COLUMN_STATUS)) == DownloadManager.STATUS_SUCCESSFUL) {
                        downloading = false;
                    }
                    //CLOSE CURSOR
                    c.close();
                    //WAIT FOR 1/4 SECOND FOR ANOTHER ITERATION
                    try {
                        Thread.sleep(250);
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            }
        }).start();
    }
    public void setSuccess(double bytes_downloaded, double bytes_total){
        if (progressSink != null) {
            progressSink.success(Arrays.asList("" + OtaStatus.DOWNLOADING.ordinal(), "" + ((bytes_downloaded * 100) / bytes_total)));
        }
    }
    private class UIhandler extends Handler {
        WeakReference<OtaUpdatePlugin> weakReference;

        public UIhandler(OtaUpdatePlugin activity) {
            this.weakReference = new WeakReference<OtaUpdatePlugin>(activity);
        }

        @Override
        public void handleMessage(Message msg) {
            super.handleMessage(msg);
            setSuccess(msg.getData().getDouble("bytes_downloaded"), msg.getData().getDouble("bytes_total"));
        }
    }
}