import 'dart:async';

import 'package:flutter/services.dart';

class OtaUpdate {
  static const EventChannel _progressChannel =
      const EventChannel('sk.fourq.ota_update');
  static Stream<OtaEvent> _progressStream;

  static Stream<OtaEvent> execute(String url, {String name}) {
    _progressStream = null;
    if (_progressStream == null) {
      _progressStream = _progressChannel.receiveBroadcastStream(
        {"url": url, "name": name},
      ).map(
        (dynamic event) => _toOtaEvent(event.cast<String>()),
      );
    }
    _progressStream.listen((OtaEvent event) {
      if(event.status == OtaStatus.INSTALLING){
        _progressStream = null;
      }
    });
    return _progressStream;
  }
  static const MethodChannel _eventChannel =
      const MethodChannel('ota_update');
  static Future callios(String url) async{
    return await _eventChannel.invokeMethod("execute", {
      "url": url,
    });
  }

  static Future exists(String name) async{
    return await _eventChannel.invokeMethod("exists", {
      "name": name,
    });
  }

  static Future installApk(String name) async{
    return await _eventChannel.invokeMethod("installApk", {
      "name": name,
    });
  }

  static OtaEvent _toOtaEvent(List<String> event) {
    return OtaEvent()
      ..status = OtaStatus.values[int.parse(event[0])]
      ..value = event[1];
  }
}

class OtaEvent {
  OtaStatus status;
  String value;
}

enum OtaStatus {
  DOWNLOADING,
  INSTALLING,
  ALREADY_RUNNING_ERROR,
  PERMISSION_NOT_GRANTED_ERROR,
  INTERNAL_ERROR
}
